#include "ma_dnp.h"
#include "utools.h"

#include <stdio.h>

/*

 credits go to UNKNOWNONE for decyphering routine and for sourcecode for his dnpextract
http://www.elitepvpers.de/forum/eo-main-discussions-questions/215944-there-working-dds-extractor.html
http://www.elitepvpers.de/forum/eo-main-discussions-questions/62447-dnp-extractor.html


I will try to describe how I do it, see if you can spot anything incorrect.
Say, I start with file name, e.g. Paladin5/effect/skill/newskill380/1.c3
1. Change to lowercase
2. Change slashes
so I end up with paladin5\effect\skill\newskill380\1.c3
Now copy it to array, then run those bit operations.

*/

char	CMassDnp::magic_string[ CMassDnp::magic_string_len + 1 ] = "DawnPack.TqDigital\x00\xCC\xCC\xCC\xCC\xCC\xCC\xCC\xCC\xCC\xCC\xCC\xCC\xCC";

char *CMassDnp::GetIdent( void )
{
	static char Ident[] = "DNP (TqDigital)";

	return Ident;
}

char *CMassDnp::GetExtension( void )
{
	static char ext[] = "dnp";
	return ext;
}

int	 CMassDnp::CheckFileType( void )
{
	if ( massf == NULL )
		return 0;
	
	char	buf[ CMassDnp::magic_string_len ];
	
	fseek( massf, 0, SEEK_SET );
	fread( buf, CMassDnp::magic_string_len, 1, massf );

	if ( memcmp( buf, magic_string, CMassDnp::magic_string_len ) == 0 )
		return 1;
	else
		return 0;
}

void CMassDnp::GetFileMainInfo( void )
{
	files_count = 0;

	if ( massf == NULL )
		return;

	fseek( massf, 0, SEEK_END );
	filesize = ftell( massf );

	fseek( massf, CMassDnp::magic_string_len + 4, SEEK_SET );
	fread( &files_count, 4, 1, massf );
	lib_start = 0x28;
	
	// read list of file names
	LoadFileNames();
}

int	 CMassDnp::ReadRec( long num )
{
	if ( massf == NULL ) 
		return 0;
	if (( (unsigned long)num >= files_count ) && ( num != -1 ))
		return 0;
	if (( num == -1 ) && 
		(( ! next_set ) || ((unsigned long)next_rec_num >= files_count )) )
		return 0;
	
	memset( &FileRec, 0, sizeof( FileRec ) );

	//seek
	if ( num == -1 ) {
		if ( fseek( massf, next_rec_pos, SEEK_SET ) != 0 )
			return 0;
	}
	else {
		pos = lib_start + 12 * num;
		if ( fseek( massf, pos, SEEK_SET ) != 0 )
			return 0;
	}

	//read
	if ( fread( &record.fileid, 4, 1, massf ) != 1 )
		return 0;
	if ( fread( &record.size, 4, 1, massf ) != 1 )
		return 0;
	if ( fread( &record.offset, 4, 1, massf ) != 1 )
		return 0;
	
	// these xors work on c3.dnp Paladin1.dnp and maybe also data.dnp which has ident 3E9
	record.size ^= 0x96120059;
	record.offset ^= 0x99589958;
	record.fileid ^= 0x95279527;

	//set position of the next
	next_rec_pos = ftell( massf );
	if ( num == -1 )
		next_rec_num ++;
	else
		next_rec_num = num + 1;
	if ( num + 1 >= (long)files_count )
		next_set = false;
	else
		next_set = true;
	
	//get info
	bool valid;
	std::string	name;
	
	name = IdToString( record.fileid, valid );
	// if filename is not found in the hash table, then find out filetype from pure file data
	if ( ! valid ) {
		
		name = "missing/" + name;
		
		char		buf[ 8 ];
		fseek( massf, record.offset, SEEK_SET );
		fread( buf, 8, 1, massf );
		if ( memcmp( buf, "DDS", 3 ) == 0 )
			name += ".dds";
		
		if ( memcmp( buf, "MAXFILE", 7 ) == 0 )
			name += ".c3";
		
	}
	
	FileRec.SetName( name.c_str() );
	FileRec.offset = record.offset;
	FileRec.size = record.size;
	FileRec.set = 1;
	
	return 1;
}

/*
unsigned long CMassDnp::StringToId2( const char *str )
{

	int i;
	unsigned int v;
	static unsigned long m[70];
	
	memcpy( m, str, strlen( str ) + 1 );
	
	i = strlen( str );
	if (( i % 4 ) == 0 )
		i /= 4;
	else
		i = i / 4 + 1;
	
	m[ i++ ] = 0x9BE74448,
	m[ i++ ] = 0x66F42C48;
	v = 0xF4FA8928;
	
	unsigned long	esi, edi, eax, ebx, ecx, edx;
	long long		ll, ll32;
	
	ll32 = 1 << 32;
	
	esi = 0x37A8470E;
	edi = 0x7758B42B;
	ecx = 0;
	
	do {
		ebx = 0x267B0B11;
		// 	rol v,1
		v = ( v << 1 ) | ( v >> 31 );
		// lea eax,m
		ebx ^= v;
		eax = m[ ecx ];
		
		edx = ebx;
		esi ^= eax;
		edi ^= eax;
		
		edx += edi;
		edx |= 0x2040801;
		edx &= 0xBFEF7FDF;
		eax = esi;
		printf( "ax: %08X bx: %08X cx: %08X dx: %08X esi: %08X edi: %08X hash: %08X\n",
			eax, ebx, ecx, edx, esi, edi, v );
		
		ll = eax;
		ll *= edx;
		eax = ll;
		
		eax += edx;
		if ( ll >= ll32 ) {
			eax++;
			eax++;
		}
		
		edx = ebx;
		edx += esi;
		edx |= 0x804021;
		edx &= 0x7DFEFBFF;
		printf( "ax: %08X bx: %08X cx: %08X dx: %08X esi: %08X edi: %08X hash: %08X\n",
			eax, ebx, ecx, edx, esi, edi, v );
		
		esi = eax;
		ll = edi;
		ll *= edx;
		eax = ll;
		
		ll = 2 * edx;
		edx = ll;
		eax += edx;
		if ( ll >= ll32 ) {
			ll = eax;
			ll++;
			eax = ll;
		} else {
			ll = 0;
		}
		
		if ( ll >= ll32 ) {
			eax += 2;
		}
		
		ecx++;
		edi = eax;
		
		printf( "ax: %08X bx: %08X cx: %08X dx: %08X esi: %08X edi: %08X hash: %08X\n",
			eax, ebx, ecx, edx, esi, edi, v );
	} while ( ecx != i );
	
	esi ^= edi;
	v = esi;
	
	printf( "hash: %08X\n", v );

	return v;
}
*/

unsigned long CMassDnp::StringToId( const char *str )
{
#ifdef _UNIX
	typedef unsigned long long t_uint64;
#else
	typedef unsigned __int64 t_uint64;
#endif

	unsigned long Hash=0xF4FA8928;
	unsigned long A=0;
	unsigned long B=0;
	unsigned long dwTemp1=0x37A8470E;
	unsigned long dwTemp2=0x7758B42B;
	unsigned long dwTemp3=0;
	t_uint64 Carrier;
	unsigned long Carry; 
	int fileNameIntCount, j=0;
	
	unsigned long *fileNameInts;
	unsigned char fileNameBytes[ 280 ];
	if ( strlen( str ) > 256 ) {
		return 0;
	}
	memset( fileNameBytes, 0, 280 );
	strcpy( (char*)fileNameBytes, str );
	fileNameInts = (unsigned long*) fileNameBytes;
	
	fileNameIntCount = strlen( str );
	if (( fileNameIntCount % 4 ) != 0 )
		fileNameIntCount = fileNameIntCount / 4 + 1;
	else
		fileNameIntCount = fileNameIntCount / 4;
	
	fileNameInts[fileNameIntCount++] = 0x9BE74448;
	fileNameInts[fileNameIntCount++] = 0x66F42C48;
	
	for (j=0;j<fileNameIntCount;j++)
	{
		dwTemp3=0x267B0B11;
		Hash = (( Hash << 1) | (Hash >> 31) );
		dwTemp3 ^= Hash;
		A=fileNameInts[j];
		dwTemp1 ^= A;
		dwTemp2 ^= A;
		B = dwTemp3;
		B += dwTemp2;
		B = B | 0x2040801;
		B = B & 0xBFEF7FDF;
		//printf( "ax: %08X bx: %08X cx: %08X dx: %08X esi: %08X edi: %08X hash: %08X\n",
		//	A, B, j, 0, dwTemp1, dwTemp2, Hash );
		Carrier = B;
		Carrier *= dwTemp1;
		A = (unsigned long)(Carrier);
		B = (unsigned long)(Carrier >> 32);
		if (B!=0) A += 1;
		Carrier = A;
		Carrier += B;
		A = (unsigned long)(Carrier);
		Carry = (unsigned long)(Carrier >> 32);
		if (Carry!=0) A += 1;
		B = dwTemp3;
		B += dwTemp1;
		B = B | 0x804021;
		B = B & 0x7DFEFBFF;
		//printf( "ax: %08X bx: %08X cx: %08X dx: %08X esi: %08X edi: %08X hash: %08X\n",
		//	A, B, j, 0, dwTemp1, dwTemp2, Hash );
		dwTemp1 = A;
		Carrier = dwTemp2;
		Carrier *= B;
		A = (unsigned long)Carrier;
		B = (unsigned long)(Carrier >> 32);
		Carrier = B;
		Carrier += B;
		B = (unsigned long)(Carrier);
		Carry = (unsigned long)(Carrier >> 32);
		if (Carry!=0) A += 1;
		Carrier = A;
		Carrier += B;
		A = (unsigned long)Carrier;
		Carry = (unsigned long)(Carrier >> 32);
		if (Carry!=0) A += 2;
		dwTemp2 = A;
	
		//printf( "ax: %08X bx: %08X cx: %08X dx: %08X esi: %08X edi: %08X hash: %08X\n",
		//	A, B, j, 0, dwTemp1, dwTemp2, Hash );
	}
	
	Hash = dwTemp1 ^ dwTemp2;
	//printf( "hash: %08X\n", Hash );
	return Hash;
}

std::string CMassDnp::IdToString( unsigned long id, bool& found )
{
	t_name_id::const_iterator	it;
	
	it = m_name_id_array.find( id );
	if ( it == m_name_id_array.end() ) {
		char	str[ 128 ];
		
		snprintf( str, 127, "%08X", id );
		found = false;
		
		return str;
	}
	
	found = true;
	return it->second;	
}

void CMassDnp::LoadFileNames( void )
{
	m_name_id_array.clear();
	
	FILE *f;
	static const char fname[] = "dnplist.txt";
	
	std::string		str1;
	
	// first try to find it in file directory
	str1 = MassFilePath;
	str1 += "/";
	str1 += fname;
	
	f = fopen( str1.c_str(), "rt" );
	if ( f == NULL ) {
		
		// next, try to find it in current directory.
		str1 = fname;
		f = fopen( str1.c_str(), "rt" );
		if ( f == NULL ) {
			UnmassTools::Log( 1, "Could not find list of filenames '%s'.\n", fname );
			return;
		}
		
	}
	
	char buf[ 512 ];
	char *cp, *cp2;
	int	i;
	
	while( ! feof( f ) ) {
		
		fgets( buf, 511, f );
		
		// remove trailing whitespace
		i = strlen( buf );
		while ( i > 0 ) {
			switch( buf[ i ] ) {
			case '\n':
			case '\r':
			case ' ':
			case '\t':
			case 0:
				i--;
				break;
			default:
				buf[ i+ 1 ] = 0;
				i = -1;
				break;
			}
		} ;
		
		// skip if line empty
		if (( i == 0 ) || ( strlen( buf ) == 0 ))
			continue;
		
		// if line is in format 'first=second' then take only 'second'
		cp = buf;
		cp2 = buf;
		while( *cp2 != 0 ) {
			if ( *cp2 == '=' ) {
				cp = cp2 + 1;
				break;
			}
			cp2++;
		}
		
		// Change to lowercase, Change slashes
		cp2 = cp;
		while( *cp2 != 0 ) {
			
			*cp2 = tolower( *cp2 );
			if ( *cp2 == '/' )
			*cp2 = '\\';
			
			cp2++;
		}
		
		// store into hash array
		m_name_id_array[ StringToId( cp ) ] = cp;
		
		// printf( "%08X = %s\n", 	StringToId( cp ), cp );
	} ;
	 
	fclose( f );
	
}

